from arlunio.doc.builder import Notebook, NotebookCell


class TestNotebook:
    """Tests for the :code:`Notebook` class."""

    def test_to_json(self):
        """Ensure that we can convert a :code:`Notebook` instance into a valid
        notebook."""

        expected = {
            "nbformat": 4,
            "nbformat_minor": 2,
            "metadata": {},
            "cells": [
                {
                    "cell_type": "code",
                    "execution_count": None,
                    "metadata": {},
                    "outputs": [],
                    "source": ["\n"],
                },
                {"cell_type": "markdown", "metadata": {}, "source": ["\n"]},
            ],
        }

        code = NotebookCell.code()
        markdown = NotebookCell.markdown()
        nb = Notebook.fromcells([code, markdown])

        assert expected == nb.json



import math

import py.test

from arlunio.imp import NotebookLoader


def test_notebook_loader_simple(testdata):
    """Ensure that we can load a simple notebook that only contains code cells"""

    nbpath = testdata("Add.ipynb", path_only=True)
    notebook = NotebookLoader.fromfile(nbpath)

    assert 3 == notebook.add(1, 2)


def test_notebook_loader(testdata):
    """Ensure that we can load a notebook that contains a mixture of cell types."""

    nbpath = testdata("Pythagoras.ipynb", path_only=True)
    notebook = NotebookLoader.fromfile(nbpath)

    assert 5 == math.sqrt(notebook.pythagoras(3, 4))


def test_notebook_loader_metadata(testdata):
    """Ensure that we can access a notebook's metadata"""

    nbpath = testdata("Metadata.ipynb", path_only=True)
    notebook = NotebookLoader.fromfile(nbpath)

    assert 6 == notebook.add(1, 2, 3)

    assert [1920, 1080] == notebook.__notebook__.metadata.arlunio.dimensions
    assert "Alex" == notebook.__notebook__.metadata.arlunio.author


def test_notebook_loader_checks_existance(testdata):
    """Ensure that the notebook loader checks to see if the given notebook exists,
    raising a helpful error if it doesn't."""

    with py.test.raises(FileNotFoundError) as err:
        NotebookLoader.fromfile("not-found.ipynb")

    assert "not-found.ipynb" in str(err.value)



import arlunio as ar
import numpy as np

from .parameters import X, Y


@ar.definition
def Circle(x: X, y: Y, *, xc=0, yc=0, r=0.8, pt=None):
    """We define a circle using the following equality.

    .. math::

       (x - x_c)^2 + (y - y_c)^2 = r^2

    where:

    - :math:`(x_0, y_0)`: Defines the centre
    - :math:`r`: Controls the radius

    In python these variables can be set using the :code:`xc`, :code:`yc`
    and :code:`r` keyword arguments to the shape's constructor.

    By default when drawn the shape will draw a filled in circle, however the
    :code:`pt` keyword argument can be used to override this. If not :code:`None`
    the value of this argument will be used to control the thickness of the line
    used to draw the circle.
    """
    x = (x - xc) ** 2
    y = (y - yc) ** 2
    circle = np.sqrt(x + y)

    if pt is None:
        return circle < r ** 2

    p = r * pt
    inner = (r - p) ** 2
    outer = (r + p) ** 2

    return ar.all(inner < circle, circle < outer)


@ar.definition
def Ellipse(x: X, y: Y, *, xc=0, yc=0, a=2, b=1, r=0.8, pt=None):
    """An ellipse can be defined using the following equality.

    .. math::

       \\left(\\frac{x - x_c}{a}\\right)^2 +
       \\left(\\frac{y - y_c}{b}\\right)^2 = r^2

    where:

    - :math:`(x_c, y_c)`: Defines the centre
    - :math:`r`: Controls the overall size
    - :math:`a`: Controls the width
    - :math:`b`: Controls the height

    In python these variables can be set using the :code:`xc`, :code:`yc`,
    :code:`r`, :code:`a` and :code:`b` keyword arguments to the shape's
    constructor.

    Increasing the value of :code:`a` will have the effect of stretching out the
    ellpise width-wise. Similarly increasing the value of :code:`b` will do the
    same but for the height. Whenever :code:`a == b` the result will be a circle.

    By default when drawn the shape will draw the ellipse filled in, this behavior
    can be changed using the :code:`pt` keyword argument. If not :code:`None` the
    value of this argument will be used as the thickness of the line used to draw
    the ellipse.
    """

    x = (x - xc) ** 2
    y = (y - yc) ** 2

    a = a ** 2
    b = b ** 2

    ellipse = np.sqrt(x / a + y / b)

    if pt is None:
        return ellipse < r * r

    p = r * pt
    inner = (r - p) ** 2
    outer = (r + p) ** 2

    return ar.all(inner < ellipse, ellipse < outer)


@ar.definition
def SuperEllipse(x: X, y: Y, *, xc=0, yc=0, a=1, b=1, n=3, r=0.8, m=None, pt=None):
    """We define a superellipse by the following equality.

    .. math::

       \\left|\\frac{(x - x_c)}{a}\\right|^n + \\left|\\frac{(y - y_c)}{b}\\right|^m = r

    where:

    - :math:`(x_c, y_c)`: Defines the center
    - :math:`r`: Controls the overall size
    - :math:`a`: Controls the width
    - :math:`b`: Controls the height
    - :math:`n`: Controls the profile of the curve far from :math:`x = 0`
    - :math:`m`: Controls the profile of the curve close to :math:`x = 0`

    Being a generalisation of an |Ellipse| the :code:`xc`, :code:`yc`, :code:`a`,
    :code:`b`, :code:`r` and :code:`pt` values behave as they would on a regular
    ellipse.

    The main difference between a `SuperEllipse`_ and a regular ellipse is the ability
    to choose the power each term on the |LHS| is raised to. This gives greater control
    over the profile of the curve that defines the shape. The power can be set
    independantly for each term through the properties :code:`n` and :code:`m`. If
    :code:`m` is :code:`None` then the shape will assume you want :math:`m = n` and set
    the value of :code:`m` accordingly.

    Considering the case where :math:`n = 2 = m`, then we recover the definition of an
    ellipse. Increasing this value, you will start seeing something that resembles a
    rectangle with rounded corners, theoretically becoming a rectangle at infinity.
    However due to limits of the implementation, the size of the shape is hard to
    control as you go much beyond :code:`10` or :code:`20`. If you want an actual
    rectangle, then perhaps the |Rectangle| shape is what you are after.

    In the range of :math:`1 \\leq n = m < 2` the "sides" of the ellipse progressively
    straighten, becoming a diamond shape at :math:`n = 1 = m`

    Finally when the values of both :math:`n,m` are less than one, the sides of the
    diamond start curving inwards towards the origin, theoretically producing something
    that looks like a :code:`+` symbol close to (but not equal to!) :code:`0`. Again
    however, due to limitations of the implementation, the shape becomes harder to
    control as you get much smaller than :code:`0.2`

    The cases where :math:`n \\neq m` the results are harder to describe! The result
    will be some combination of the results described above, but you are probably better
    off experimenting and seeing what comes out of it!

    .. _SuperEllipse: https://en.wikipedia.org/wiki/Superellipse

    """

    x = x - xc
    y = y - yc

    if m is None:
        m = n

    ellipse = np.abs(x / a) ** n + np.abs(y / b) ** m

    if pt is None:
        return ellipse < r

    p = r * pt
    inner = r - p
    outer = r + p

    return ar.all(inner < ellipse, ellipse < outer)


@ar.definition
def Square(x: X, y: Y, *, xc=0, yc=0, size=0.8, pt=None):
    """A square."""

    xs = np.abs(x)
    ys = np.abs(y)

    if pt is None:
        return ar.all(xs < size, ys < size)

    inner = ar.all(xs < size - pt, ys < size - pt)
    outer = ar.all(xs < size + pt, ys < size + pt)

    return ar.all(outer, ar.invert(inner))


@ar.definition
def Rectangle(x: X, y: Y, *, xc=0, yc=0, size=0.6, ratio=1.618, pt=None):
    """A Rectangle."""

    xs = np.abs(x)
    ys = np.abs(y)
    width = size * ratio
    height = size

    if pt is None:
        return ar.all(xs < width, ys < height)

    inner = ar.all(xs < width - pt, ys < height - pt)
    outer = ar.all(xs < width + pt, ys < height + pt)

    return ar.all(outer, ar.invert(inner))



import logging

import arlunio as ar

from .components import ImageViewer

try:
    import tkinter as tk
    import tkinter.ttk as ttk

    TK = True
except ImportError:
    import unittest.mock as mock

    tk = mock.MagicMock()
    ttk = mock.MagicMock()

    TK = False


logger = logging.getLogger(__name__)


class ShapeDesigner(tk.Frame):
    """UI for designing new shapes"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.pack(fill=tk.BOTH, expand=tk.TRUE)

        self.viewer = ImageViewer(self)
        self.viewer.pack(fill=tk.BOTH, expand=tk.TRUE, side=tk.LEFT)
        self.viewer.set_item(ar.S.Circle())
        self.shapes = {s.__name__: s for s in ar.S._items.values()}

        self.shape_picker = ttk.Combobox(self, values=list(self.shapes.keys()))
        self.shape_picker.pack(side=tk.TOP)
        self.shape_picker.current(0)
        self.shape_picker.bind("<<ComboboxSelected>>", self.render_shape)

    def render_shape(self, event):
        logger.debug(event)
        name = self.shape_picker.get()
        shape = self.shapes[name]()

        self.viewer.set_item(shape)


class Shapes:
    """Launches a simple shape previewer."""

    def run(self):

        if not TK:
            logger.info(
                "It appears that tkinter is not available on your system but "
                "is required for this command. Please ensure it is available "
                "and try again "
            )
            return 1

        root = tk.Tk()
        app = ShapeDesigner(root)
        app.mainloop()



import arlunio as ar
import numpy as np


@ar.definition
def X(width, height, *, x0=0, scale=1, stretch=False):
    """Cartesian :math:`x` coordinates.

    :param x0: Shift the :math:`x` coordinates by :code:`x0`
    :param scale: Controls the size of the extreme values in the grid.
    :param stretch: If :code:`True` and the image is wider than it is tall then the grid
                    will be stretched so that :code:`x = scale` falls on the image
                    border.
    """
    ratio = width / height

    if not stretch and ratio > 1:
        scale = scale * ratio

    x = np.linspace(-scale, scale, width)
    x = np.array([x for _ in range(height)])

    return x - x0


@ar.definition
def Y(width, height, *, y0=0, scale=1, stretch=False):
    """Cartesian :math:`y` coordinates.

    :param y0: Shift the :math:`y` coordinates by :code:`y0`
    :param scale: Controls the size of the extreme values in the grid
    :param stretch: If :code:`True` and the image is taller than it is wide then the
                    grid will be stretched so that :code:`y = scale` falls on the image
                    border
    """
    ratio = height / width

    if not stretch and ratio > 1:
        scale = scale * ratio

    y = np.linspace(scale, -scale, height)
    y = np.array([y for _ in range(width)]).transpose()

    return y - y0


@ar.definition
def R(x: X, y: Y):
    """Polar :math:`r` coordinates."""
    return np.sqrt(x * x + y * y)


@ar.definition
def T(x: X, y: Y, *, t0=0):
    """Polar, :math:`t` coordinates."""
    t = np.arctan2(y, x)
    return t - t0



import base64
import enum
import io
import logging
import pathlib

import numpy as np
import PIL.Image as Image
import PIL.ImageColor as Color

from ._expressions import lerp

logger = logging.getLogger(__name__)


class Resolutions(enum.Enum):
    """Enum that defines some common image resolutions

    Members of this enum are tuples containing the width and height which can be
    accessed by name::

       >>> from arlunio import Resolutions as R

       >>> hd = R.HD
       >>> hd.width
       1280

       >>> hd.height
       720

    Resolutions can also unpacked::

       >>> width, height = hd
       >>> width
       1280

       >>> height
       720
    """

    HD = (1280, 720)
    """1280 x 720"""

    FHD = (1920, 1080)
    """1920 x 1080"""

    QHD = (2560, 1440)
    """2560 x 1440"""

    def __iter__(self):
        value = self.value
        return iter([value[0], value[1]])

    @property
    def width(self):
        return self.value[0]

    @property
    def height(self):
        return self.value[1]


def save(image, filename: str, mkdirs: bool = False) -> None:
    """Save an image in PNG format.

    :param filename: The filepath to save the image to.
    :param mkdirs: If true, make any parent directories
    """
    path = pathlib.Path(filename)

    if not path.parent.exists() and mkdirs:
        path.parent.mkdir(parents=True)

    with open(filename, "wb") as f:
        image.save(f)


def encode(image) -> bytes:
    """Return the image encoded as a base64 string."""

    with io.BytesIO() as byte_stream:
        image.save(byte_stream, "PNG")
        image_bytes = byte_stream.getvalue()

        return base64.b64encode(image_bytes)


def colorramp(values, start=None, stop=None):
    """Given a range of values, produce an image mapping those values onto colors."""

    (r, g, b) = Color.getrgb("#000") if start is None else Color.getrgb(start)
    (R, G, B) = Color.getrgb("#fff") if stop is None else Color.getrgb(stop)

    reds = np.floor(lerp(r, R)(values))
    greens = np.floor(lerp(g, G)(values))
    blues = np.floor(lerp(b, B)(values))

    pixels = np.array(np.dstack([reds, greens, blues]), dtype=np.uint8)
    return Image.fromarray(pixels)


def fill(mask, color=None, background=None):
    """Given a mask, fill it in with a color."""

    color = "#000" if color is None else color
    background = "#fff" if background is None else background

    mask_img = Image.fromarray(mask)
    fill_color = Color.getrgb(color)

    height, width = mask.shape

    image = Image.new("RGB", (width, height), color=background)
    image.paste(fill_color, mask=mask_img)

    return image



import typing

import arlunio

from sphinx.application import Sphinx

from .builder import NotebookTutorialBuilder
from .directives import (
    ArlunioImageDirective,
    AutoDefinitionDirective,
    NBTutorialDirective,
    depart_nbtutorial,
    nbtutorial,
    visit_nbtutorial,
)


def setup(app: Sphinx) -> typing.Dict[str, typing.Any]:

    app.add_node(
        nbtutorial,
        html=(visit_nbtutorial, depart_nbtutorial),
        latex=(visit_nbtutorial, depart_nbtutorial),
        text=(visit_nbtutorial, depart_nbtutorial),
    )
    app.add_builder(NotebookTutorialBuilder)

    app.add_directive("autodefn", AutoDefinitionDirective)
    app.add_directive("nbtutorial", NBTutorialDirective)
    app.add_directive("arlunio-image", ArlunioImageDirective)

    return {"version": arlunio.__version__, "parallel_read_safe": True}



import logging

import arlunio as ar
import numpy as np

logger = logging.getLogger(__name__)


@ar.definition
def Grid(width, height, *, nx=4, ny=4, shape=None):
    """Repeatedly draw the given shape in a grid.

    :param nx: The number of times to repeat the shape in the x-direction
    :param ny: The number of times to repeat the shape in the y-direction
    :param shape: The shape instance to draw.
    """
    bg = np.full((height, width), False)

    logger.debug(f"Grid size: {ny} x {nx}")
    logger.debug(f"Background size: {height} x {width}")

    # Draw the shape at a size determined by the size of the grid
    s_height, s_width = height // ny, width // nx
    mask = shape(s_width, s_height)

    logger.debug(f"Shape size: {s_height} x {s_width}")

    # Let numpy handle the repeating of the shape across the image.
    pattern = np.tile(mask, (ny, nx))

    # Apply the pattern to the background, depending on the grid size and
    # image dimensions align, the generated grid may not perfectly fill the
    # image.
    p_height, p_width = pattern.shape
    logger.debug(f"Pattern size: {p_height} x {p_width}")

    bg[:p_height, :p_width] = pattern

    return bg


@ar.definition
def Map(width, height, *, layout=None, legend=None):
    """For more complex layouts."""

    nx, ny = len(layout), len(layout[0])
    size = (height // ny, width // nx)  # TODO: Handle divisions with rounding errors

    # Build a new dict with the values being the shapes drawn at the appropriate res
    # to ensure we only draw them once.
    items = {k: v(*size) for k, v in legend.items()}
    return np.block([[items[key] for key in row] for row in layout])


@ar.definition
def Pixelize(width, height, *, pixels=None, shape=None, nx=None, ny=None):
    """Draw a pixelated version of a shape."""

    if shape is None and pixels is None:
        raise ValueError("You must either provide a shape or a pixel pattern.")

    if shape is not None:

        if nx is None or ny is None:
            raise ValueError("You must also provide the `nx` and `ny` size parameters")

        pixels = shape(nx, ny)

    nx, ny = len(pixels), len(pixels[0])
    size = (height // ny, width // nx)  # TODO: Handle divisions with rounding errors

    fill = np.full(size, True)
    empty = np.full(size, False)

    return np.block([[fill if col else empty for col in row] for row in pixels])



import textwrap

from arlunio.doc.directives import format_error, render_image
from docutils import nodes
from docutils.statemachine import StringList


def test_render_image_handles_syntax_error():
    """Ensure that if there is an error when compiling the code we can
    handle it."""

    src = "x 1 + 2"
    doctree = render_image(src)

    assert len(doctree) == 1
    assert isinstance(doctree[0], nodes.error), "Expected error node."

    error = doctree[0]
    assert len(error.children) == 2, "Expected two content nodes."

    text, trace = error.children

    assert isinstance(text, nodes.Text), "Expected text node."
    assert "Unable to render image" in text.astext()

    assert isinstance(trace, nodes.literal_block)
    assert "Traceback" in trace.astext()
    assert "SyntaxError" in trace.astext()


def test_render_image_handles_runtime_error():
    """Ensure that if there is an error when executing the code we can handle it."""

    src = "x = 1 / 0"
    doctree = render_image(src)

    assert len(doctree) == 1
    assert isinstance(doctree[0], nodes.error), "Expected error node."

    error = doctree[0]
    assert len(error.children) == 2, "Expected two content nodes."

    text, trace = error.children

    assert isinstance(text, nodes.Text), "Expected text node."
    assert "Unable to render image" in text.astext()

    assert isinstance(trace, nodes.literal_block)
    assert "Traceback" in trace.astext()
    assert "ZeroDivisionError" in trace.astext()


def test_render_image_image_provided():
    """Ensure that if the code provides an image we use that."""

    src = """\
    import arlunio as ar
    from arlunio.lib.shapes import Circle

    circle = Circle()
    disk = ar.fill(circle(4,4))
    """
    doctree = render_image(textwrap.dedent(src))

    assert len(doctree) == 1
    assert isinstance(doctree[0], nodes.raw), "Expected raw node"


def test_format_error():

    err = "\n".join(
        [
            "Traceback (most recent call last):",
            '   File "<stdin>", line 1, in <module>',
            "ZeroDivisionError: division by zero",
        ]
    )
    content = format_error("Unable to perform :code:`2 / 0`", err)

    assert isinstance(content, StringList)

    actual = list(content)
    expected = [
        ".. error::",
        "",
        "   Unable to perform :code:`2 / 0`::",
        "",
        "      Traceback (most recent call last):",
        '         File "<stdin>", line 1, in <module>',
        "      ZeroDivisionError: division by zero",
        "",
    ]

    assert expected == actual



import importlib
import string
import textwrap
import traceback

from typing import List

import arlunio as ar
import attr

from docutils import nodes
from docutils.parsers import rst
from docutils.statemachine import StringList
from PIL.Image import Image
from sphinx.util import logging, nested_parse_with_titles

logger = logging.getLogger(__name__)


ERROR_TEMPLATE = """\
.. error::

${message}

${traceback}
"""

DEFINITION_TEMPLATE = """\
${definition_name}

.. py:class:: ${definition_path}

${definition_image}

${definition_desc}

${definition_props}
"""

ATTRIBUTES_TEMPLATE = """\
:Attributes:

${attrs}
"""

ATTRIBUTE_TEMPLATE = """\
- **${name}** ${type} - Default: :code:`${default}`
"""

PREVIEW_TEMPLATE = """\
.. raw:: html

   <figure>
     <img alt="${alt_tag}"
          src="data:image/png;base64,${img_data}"
          style="border: solid 1px #ddd"/>
     <figcaption style="margin: 10px;text-align:center">
       <small style="font-style:italic">${alt_tag}</small>
     </figcaption>
   </figure>

"""

IMAGE_TEMPLATE = string.Template(
    """\
<figure style="border: solid 1px #ddd;width: 75%;margin:auto">
  <img style="image-rendering:crisp-edges;width:100%"
       src="data:image/png;base64,$data"></img>
</figure>
"""
)


class nbtutorial(nodes.General, nodes.Element):
    pass


def visit_nbtutorial(self, node):
    pass


def depart_nbtutorial(self, node):
    pass


def load_definition(object_spec: str) -> (ar.Definition, str):
    """Given a classpath e.g. :code:`arlunio.lib.basic.Circle` load it.

    There is an issue(?) currently with shapes defined with the :code:`@shape` decorator
    where the shape's module is reported as the module where the :code:`@shape`
    decorator is defined (:code:`arlunio._shapes`) rather than the module where the
    shape is defined (:code:`arlunio.lib.basic`). So this function also returns the true
    module name::

       >>> from arlunio.doc.directives import load_definition

       >>> load_definition("arlunio.lib.shapes.Circle")
       (<class 'arlunio._core.Circle'>, 'arlunio.lib.shapes')

    If the module is not found then a :code:`MoudleNotFoundError` will be raised::

       >>> load_definition("arlunio.notfound.Circle")
       Traceback (most recent call last):
          ...
       ModuleNotFoundError: No module named 'arlunio.notfound'

    If the class within the module cannot be found then an :code:`AttributeError` will
    be raised::

       >>> load_definition("arlunio._core.Circle")
       Traceback (most recent call last):
          ...
       AttributeError: module 'arlunio._core' has no attribute 'Circle'

    Finally if the given class is not a shape then a :code:`TypeError` will be raised::

       >>> load_definition("arlunio.doc.directives.AutoDefinitionDirective")
       Traceback (most recent call last):
          ...
       TypeError: 'arlunio.doc.directives.AutoDefinitionDirective' is not a definition
    """

    *obj_path, obj_name = object_spec.split(".")
    obj_path = ".".join(obj_path)

    logger.debug(f"[autodefn]: Looking for definition: {obj_name} in {obj_path}")

    module = importlib.import_module(obj_path)
    defn = getattr(module, obj_name)

    if not issubclass(defn, ar.Definition):
        raise TypeError(f"'{object_spec}' is not a definition")

    return defn, module.__name__


def document_attributes(defn: ar.Definition) -> str:
    """Given a definition, document all of its properties."""
    template = string.Template(ATTRIBUTES_TEMPLATE)
    attr_tpl = string.Template(ATTRIBUTE_TEMPLATE)

    attributes = []

    for attrib in sorted(attr.fields(defn), key=lambda p: p.name):

        type_ = "" if attrib.type is None else f"({str(attrib.type)})"
        context = {"name": attrib.name, "default": attrib.default, "type": type_}

        attributes.append(attr_tpl.safe_substitute(context))

    return template.safe_substitute({"attrs": "\n".join(attributes)})


def generate_preview(defn: ar.Definition) -> str:
    """Given an instance of the shape, generate a preview of it."""

    # TODO: Figure out how to generalise this for definitions...
    # template = string.Template(PREVIEW_TEMPLATE)
    # image = shape_ins(1920, 1080)
    # name = shape_ins.__class__.__name__

    # context = {
    #    "img_data": image.encode().decode("utf-8"),
    #    "alt_tag": f"Example instance of the {name} shape",
    # }

    return ""  # template.safe_substitute(context)


def document_definition(defn: ar.Definition, module_name: str) -> StringList:
    """Given a definition, automatically write the reference documentation for it.
    """
    default = defn()
    indent = " " * 3
    template = string.Template(DEFINITION_TEMPLATE)
    docstring = "" if defn.__doc__ is None else defn.__doc__

    context = {
        "definition_name": defn.__name__ + "\n" + "-" * len(defn.__name__),
        "definition_path": f"{module_name}.{defn.__name__}",
        "definition_desc": textwrap.indent(docstring, indent),
        "definition_props": textwrap.indent(document_attributes(defn), indent),
        "definition_image": textwrap.indent(generate_preview(default), indent),
    }

    documentation = template.safe_substitute(context)

    return StringList(documentation.split("\n"), source="")


def format_error(message: str, err: str) -> StringList:
    """Given an error message format it as an error."""

    indent = " " * 3
    msg = textwrap.indent(message + "::", indent)
    err_msg = textwrap.indent(err, indent * 2)

    template = string.Template(ERROR_TEMPLATE)
    error = template.safe_substitute({"message": msg, "traceback": err_msg})

    return StringList(error.split("\n"), source="")


def parse_content(state, content: StringList) -> List[nodes.Node]:
    """Given a reStructured text representation of some content format parse it
    into a list of nodes."""

    section = nodes.section()
    section.document = state.document
    nested_parse_with_titles(state, content, section)

    return section.children


def render_image(src: str) -> List[nodes.Node]:
    """Given the source code for an image return a doctree that when rendered by
    Sphinx will insert that image into a HTML page.

    :param src: The source code that produces the image.
    """
    doctree = []

    try:
        code = compile(src, "<string>", "exec")
    except Exception:
        message = nodes.Text("Unable to render image: Invalid code")
        err = nodes.literal_block("", traceback.format_exc())
        doctree.append(nodes.error("", message, err))

        return doctree

    environment = {}

    try:
        exec(code, environment)
    except Exception:
        message = nodes.Text("Unable to render image: Error in code")
        err = nodes.literal_block("", traceback.format_exc())
        doctree.append(nodes.error("", message, err))

        return doctree

    # Look to see if the code produced an image object for us.
    image = None

    for obj in environment.values():
        if isinstance(obj, Image):
            image = obj

    if image is not None:
        context = {"data": ar.encode(image).decode("utf-8")}
        html = IMAGE_TEMPLATE.safe_substitute(context)
        doctree.append(nodes.raw("", html, format="html"))

    return doctree


class AutoDefinitionDirective(rst.Directive):
    """Given the a definition automatically generate its documentation."""

    required_arguments = 1
    optional_arguments = 0
    add_index = False

    def run(self) -> List[nodes.Node]:

        defn_name = self.arguments[0]

        try:
            defn, module_name = load_definition(defn_name)

        except Exception:
            err = traceback.format_exc()
            message = f"**AutoDefinition:** Unable to load shape: :code:`{defn_name}`"
            content = format_error(message, err)

            return parse_content(self.state, content)

        content = document_definition(defn, module_name)

        section = nodes.section()
        section.document = self.state.document
        nested_parse_with_titles(self.state, content, section)

        return section.children


class NBTutorialDirective(rst.Directive):
    def run(self):
        return [nbtutorial("")]


class ArlunioImageDirective(rst.Directive):
    """Given some code that produces an image, render it in the page."""

    has_content = True

    def run(self):

        src = "\n".join(self.content)
        return render_image(src)



import logging
import os
import shutil
import subprocess

import pkg_resources

import appdirs

logger = logging.getLogger(__name__)


class Tutorial:
    """Launch the interactive tutorial.

    This will launch a jupyter-lab instance in a folder containing a collection
    of jupyter notebook files that contain the tutorial. These files are yours
    and you are free to edit them as you see fit.

    You can also reset the tutorial at any time by launching the tutorial
    with the --reset flag. WARNING: This will delete any changes you have
    made to the tutorial directory, be sure to back up anything you wish to
    preserve BEFORE running the command with this flag

    :param reset: reset the tutorial to its default state
    """

    def _copy_resources(self, destination):
        """Copy the tutorial resources to the given destination."""
        logger.debug(f"Copying tutorial resources to {destination}")
        os.makedirs(destination)

        def exclude_item(item, path):
            return any([item.startswith("."), "__" in item])

        for item in pkg_resources.resource_listdir("arlunio.tutorial", "."):
            path = pkg_resources.resource_filename("arlunio.tutorial", item)

            if exclude_item(item, path):
                logger.debug(f"--> {item}, skipping")
                continue

            logger.debug(f"--> {item}")
            dest = os.path.join(destination, item)

            if os.path.isdir(path):
                shutil.copytree(path, dest, copy_function=shutil.copy)
            else:
                shutil.copy(path, dest)

    def run(self, reset: bool = False):
        tutorial_dir = os.path.join(
            appdirs.user_data_dir(appname="arlunio", appauthor="swyddfa"), "tutorial"
        )

        if reset and os.path.exists(tutorial_dir):
            logger.info("Existing tutorial found found, resetting...")
            shutil.rmtree(tutorial_dir)

        if not os.path.exists(tutorial_dir):
            logger.info("Copying tutorial resources...")
            self._copy_resources(tutorial_dir)

        subprocess.run(["jupyter-lab"], cwd=tutorial_dir)



import functools

from typing import Callable, Union

import numpy as np


def any(*args: Union[bool, np.ndarray]) -> Union[bool, np.ndarray]:
    """Given a number of conditions, return :code:`True` if any of the conditions
    are true.

    This function is implemented as a thin wrapper around numpy's
    :code:`np.logical_or` function so that it can take an arbitrary number of inputs.
    This also means that this function will accept arrays of differing sizes - assuming
    that they can be broadcasted to a common shape.

    .. seealso::

       |numpy.Broadcasting|
          Numpy documentation on broadcasting.

       |numpy.Array Broadcasting|
          Further background on broadcasting.

       |numpy.logical_or|
          Reference documentation on the :code:`np.logical_or` function

    :param args: A number of boolean conditions, a condition can either be a single
                 boolean value or a numpy array of boolean values.

    :Examples:

    >>> import arlunio as ar
    >>> ar.any(True, False, False)
    True
    >>> ar.any(False, False, False, False)
    False

    If the arguments are boolean numpy arrays, then the any condition is applied
    element-wise

    >>> import numpy as np
    >>> x1 = np.array([True, False, True])
    >>> x2 = np.array([False, False, True])
    >>> x3 = np.array([False, True, False])
    >>> ar.any(x1, x2, x3)
    array([ True,  True,  True])

    This function can even handle a mixture of arrays and single values - assuming
    their shapes can be broadcasted to a common shape.

    >>> ar.any(False, np.array([True, False]), np.array([[False, True], [True, False]]))
    array([[ True,  True],
           [ True, False]])
    """
    return functools.reduce(np.logical_or, args)


def all(*args: Union[bool, np.ndarray]) -> Union[bool, np.ndarray]:
    """Given a number of conditions, return :code:`True` only if **all**
    of the given conditions are true.

    This function is implemented as a thin wrapper around numpy's
    :code:`np.logical_and` function so that it can take an arbitrary number of inputs.
    This also means that this function will accept arrays of differing sizes - assuming
    that they can be broadcasted to a common shape.

    .. seealso::

       |numpy.Broadcasting|
          Numpy documentation on broadcasting.

       |numpy.Array Broadcasting|
          Further background on broadcasting.

       |numpy.logical_and|
          Reference documentation on the :code:`logical_and` function.

    :param args: A number of boolean conditions, a conditon can either be a single
                 boolean value, or a numpy array of boolean values.

    :Examples:

    >>> import arlunio as ar
    >>> ar.all(True, True, True)
    True
    >>> ar.all(True, False, True, True)
    False

    If the arguments are boolean numpy arrays, then the any condition is applied
    element-wise

    >>> import numpy as np
    >>> x1 = np.array([True, False, True])
    >>> x2 = np.array([False, False, True])
    >>> x3 = np.array([False, True, True])
    >>> ar.all(x1, x2, x3)
    array([False, False,  True])

    This function can even handle a mixture of arrays and single values - assuming
    their shapes can be broadcasted to a common shape.

    >>> ar.all(True, np.array([True, False]), np.array([[False, True], [True, False]]))
    array([[False, False],
           [ True, False]])
    """
    return functools.reduce(np.logical_and, args)


def clamp(values, min_value=0, max_value=1):
    """Force an array of values to stay within a range of values.

    :param values: The array of values to clamp
    :param min_value: The minimum value the result should contain (Default :code:`0`)
    :param max_value: The maximum value the resul should contain (Default :code:`1`)

    :Examples:

    By default values will be limited to between :code:`0` and :code:`1`

    >>> import arlunio as ar
    >>> import numpy as np
    >>> vs = np.linspace(-1, 2, 6)
    >>> ar.clamp(vs)
    array([0. , 0. , 0.2, 0.8, 1. , 1. ])

    But this can be changed with extra arguments to the :code:`clamp` function

    >>> ar.clamp(vs, min_value=-1, max_value=0.5)
    array([-1. , -0.4,  0.2,  0.5,  0.5,  0.5])
    """
    vs = np.array(values)
    vs[vs > max_value] = max_value
    vs[vs < min_value] = min_value

    return vs


def invert(x):
    return np.logical_not(x)


def lerp(start: float = 0, stop: float = 1) -> Callable[[float], float]:
    """Return a function that will linerarly interpolate between a and b.

    :param start: The value the interpolation should start from. (Default :code:`0`)
    :param stop: The value the interpolation should stop at. (Default :code:`1`)

    :Examples:

    By default this function will interpolate between :code:`0` and :code:`1`

    >>> import arlunio as ar
    >>> f = ar.lerp()
    >>> f(0)
    0
    >>> f(1)
    1

    However by passing arguments to the :code:`lerp` function we can change the bounds
    of the interpolation.

    >>> import numpy as np
    >>> ts = np.linspace(0, 1, 4)
    >>> f = lerp(start=3, stop=-1)
    >>> f(ts)
    array([ 3.        ,  1.66666667,  0.33333333, -1.        ])

    """

    def f(t: float) -> float:
        return (1 - t) * start + t * stop

    return f



import inspect

from typing import List

import py.test

from arlunio.cli._interface import CliCommand, CliOption


def mkparam(name, type_, default=None):
    """Construct parameter instances for test purposes."""
    pos = inspect.Parameter.POSITIONAL_OR_KEYWORD
    value = default if default is not None else inspect.Parameter.empty

    return inspect.Parameter(name, pos, default=value, annotation=type_)


BOOL_STORE_TRUE = {"action": "store_true"}
BOOL_STORE_FALSE = {"action": "store_false"}
INT_NO_DEFAULT = {"type": int, "default": None}
INT_WITH_DEFAULT = lambda v: {"type": int, "default": v}  # noqa: E731
FLOAT_NO_DEFAULT = {"type": float, "default": None}
FLOAT_WITH_DEFAULT = lambda v: {"type": float, "default": v}  # noqa: E731


@py.test.mark.parametrize(
    "parameter,expected",
    [
        (mkparam("reset", bool), CliOption(name="--reset", args=BOOL_STORE_TRUE)),
        (mkparam("num_cats", int), CliOption("--num-cats", args=INT_NO_DEFAULT)),
        (mkparam("radius", float), CliOption("--radius", args=FLOAT_NO_DEFAULT)),
        (
            mkparam("reset", bool, default=False),
            CliOption("--reset", args=BOOL_STORE_TRUE),
        ),
        (
            mkparam("reset", bool, default=True),
            CliOption("--reset", args=BOOL_STORE_FALSE),
        ),
        (
            mkparam("num_cats", int, default=12),
            CliOption("--num-cats", args=INT_WITH_DEFAULT(12)),
        ),
        (
            mkparam("radius", float, default=3.14),
            CliOption(name="--radius", args=FLOAT_WITH_DEFAULT(3.14)),
        ),
    ],
)
def test_cli_option_from_std_types(parameter, expected):
    """Ensure we can construct a CliOption based on the information we get
    from type annotations for standard types."""

    opt = CliOption.fromparam(parameter)
    assert opt == expected


def test_cli_option_unexpected_type():
    """Ensure that we raise an appropriate exception if we are given a type
    we cannot handle."""

    with py.test.raises(TypeError) as err:
        param = mkparam("collection", List[int])
        CliOption.fromparam(param)

    assert "Unsupported cli option type" in str(err.value)
    assert "List" in str(err.value)


def simple_func(x: int, y: float = 0.5):
    pass


def class_method(self, reset: bool, num_cats: int = 3):
    pass


@py.test.mark.parametrize(
    "method,expected",
    [
        (
            simple_func,
            [
                CliOption(name="--x", args=INT_NO_DEFAULT),
                CliOption(name="--y", args=FLOAT_WITH_DEFAULT(0.5)),
            ],
        ),
        (
            class_method,
            [
                CliOption(name="--reset", args=BOOL_STORE_TRUE),
                CliOption(name="--num-cats", args=INT_WITH_DEFAULT(3)),
            ],
        ),
    ],
)
def test_cli_option_from_method(method, expected):
    """Ensure we can construct a list of CliOptions from a given method."""

    opts = CliOption.frommethod(method)
    assert opts == expected


def test_single_cmd_no_run_method():
    """Ensure we raise an appropriate error when given a class with no run method."""

    class BadCommand:
        pass

    with py.test.raises(TypeError) as err:
        CliCommand.fromcmd("test", BadCommand)

    assert "Missing expected method" in str(err.value)
    assert "run" in str(err.value)


def test_single_cmd_run_attribute():
    """Ensure we raise an appropriate error when given a class whose 'run' attribure is
    not callable"""

    class BadCommand:
        run = 23

    with py.test.raises(TypeError) as err:
        CliCommand.fromcmd("test", BadCommand)

    assert "Missing expected method" in str(err.value)
    assert "run" in str(err.value)


class AddCommand:
    def run(self, x: int, y: int = 1):
        pass


ADD_COMMAND = CliCommand(
    name="test",
    options=[
        CliOption(name="--x", args=INT_NO_DEFAULT),
        CliOption(name="--y", args=INT_WITH_DEFAULT(1)),
    ],
)


@py.test.mark.parametrize("definition,expected", [(AddCommand, ADD_COMMAND)])
def test_single_cmd(definition, expected):
    """Ensure we can construct appropriate command defintions from a given class"""

    cmd = CliCommand.fromcmd("test", definition)
    assert cmd == expected



import arlunio.testing as T
import numpy as np
import numpy.testing as npt
import py.test

from arlunio.lib.parameters import X, Y
from hypothesis import assume, given


@given(width=T.dimension, height=T.dimension)
def test_X_matches_dimension(width, height):
    """Ensure that the output shape matches the width and height of the image"""

    x = X()
    assert x(width, height).shape == (height, width)


@given(width=T.dimension, height=T.dimension)
def test_X_uniform_vertically(width, height):
    """Ensure that the output only varies in the x-direction"""

    x = X()
    xs = x(width, height)

    assert np.unique(xs, axis=0).shape == (1, width)


@given(width=T.dimension, height=T.dimension, scale=T.pve_num)
def test_X_adapts_to_image_ratio_by_default(width, height, scale):
    """Ensure that the output adapts to the aspect ratio of the image."""

    assume(width / height > 1)

    x = X(scale=scale)
    xs = x(width, height)

    assert width / height == py.test.approx(np.max(xs) / scale)


@given(width=T.dimension, height=T.dimension, scale=T.pve_num)
def test_X_fits_to_image_size_when_told(width, height, scale):
    """Ensure that the output fits to the size of the image when :code:`fit` property
    is set."""

    x = X(scale=scale, stretch=True)
    xs = x(width, height)

    assert np.max(xs) == scale


@given(width=T.dimension, height=T.dimension, offset=T.real_num)
def test_X_shifts_origin_accordingly(width, height, offset):
    """Ensure that the output has shifted the output according to the :code:`x0`
    property."""

    x1 = X()
    x2 = X(x0=offset)

    x1s = x1(width, height)
    x2s = x2(width, height)

    npt.assert_almost_equal(x1s - x2s, offset)


@given(width=T.dimension, height=T.dimension)
def test_Y_matches_dimension(width, height):
    """Ensure that the output shape matches the width and height of the image."""

    y = Y()
    ys = y(width, height)

    assert ys.shape == (height, width)


@given(width=T.dimension, height=T.dimension)
def test_Y_uniform_horizontally(width, height):
    """Ensure that the output only varies in the y direction"""

    y = Y()
    ys = y(width, height)

    assert np.unique(ys, axis=1).shape == (height, 1)


@given(width=T.dimension, height=T.dimension, scale=T.pve_num)
def test_Y_adapts_to_image_ratio_by_default(width, height, scale):
    """Ensure that the output adapts to the aspect ratio of the image"""

    assume(height / width > 1)

    y = Y(scale=scale)
    ys = y(width, height)

    assert height / width == py.test.approx(np.max(ys) / scale)


@given(width=T.dimension, height=T.dimension, scale=T.pve_num)
def test_Y_fits_to_image_size_when_told(width, height, scale):
    """Ensure that the output fits to the size of the image when :code:`fit` property
    is set."""

    y = Y(scale=scale, stretch=True)
    ys = y(width, height)

    assert np.max(ys) == scale


@given(width=T.dimension, height=T.dimension, offset=T.real_num)
def test_Y_shifts_origin_accordingly(width, height, offset):
    """Ensure that the output has shifted the output according to the :code:`y0`
    property."""

    y1 = Y()
    y2 = Y(y0=offset)

    y1s = y1(width, height)
    y2s = y2(width, height)

    npt.assert_almost_equal(y1s - y2s, offset)



import inspect

from typing import Any, ClassVar, Dict, List

import attr


class Key:
    """This class is used to represent x.y.z keys in a collection.

    It's only a thin wrapper around a tuple that should make the implementation of
    collections easier to reason about. A key can be created from a number of strings

    >>> from arlunio._core import Key
    >>> Key('a', 'b', 'c')
    k'a.b.c'

    Alternatively from a single string with each part delimited with a dot.

    >>> Key.fromstring('a.b.c')
    k'a.b.c'

    Keys have a length which is equal to the number of components that makes up the
    key.

    >>> k = Key('a', 'b', 'c')
    >>> len(k)
    3

    They can be checked to see if they are equal against other keys, or their string
    representation

    >>> k == Key.fromstring('a.b.c')
    True

    >>> k == "a.b.c"
    True

    Keys can be indexed

    >>> k[0]
    'a'

    But not mutated

    >>> k[1] = 'd'
    Traceback (most recent call last):
        ...
    TypeError: 'Key' object does not support item assignment
    """

    __slots__ = "_key"

    def __init__(self, *args):
        self._key = args

    def __repr__(self):
        return f"k'{str(self)}'"

    def __str__(self):
        return ".".join(self._key)

    def __hash__(self):
        return hash(self._key)

    def __eq__(self, other):

        if isinstance(other, Key):
            return self._key == other._key

        if isinstance(other, str):
            return str(self) == other

        return False

    def __len__(self):
        return len(self._key)

    def __getitem__(self, index):
        return self._key[index]

    def __add__(self, other):
        return Key.fromstring(str(self) + "." + str(other))

    @classmethod
    def fromstring(cls, string):
        return cls(*string.split("."))


@attr.s(auto_attribs=True)
class Collection:
    """A group of related items, indexed by keys."""

    _items: Dict[Key, Any] = attr.Factory(dict)

    def __str__(self):
        sep = "\n|  "
        name = self.__class__.__name__
        header = f"{name}: {len(self)} items{sep}"
        items = sep.join(str(k) for k in self._items.keys())

        return header + items

    def __len__(self):
        return len(self._items)

    def __getitem__(self, key):

        if isinstance(key, int):
            return list(self._items.values())[key]

        raise KeyError(key)

    def __getattr__(self, name):
        candidates = self.find(name)
        cls = self.__class__

        if len(candidates) == 0:
            items = {k: v for k, v in self._items.items() if k[0] == name}

            if len(items) == 0:
                raise AttributeError(f"No item with name: {name}")

            return cls(items=items)

        if len(candidates) == 1:
            return candidates[0]

        raise AttributeError(f"Ambiguous reference: {name}")

    def find(self, name: str):
        """Return a collection of all the items that have the given name."""

        items = {k: v for k, v in self._items.items() if k[-1] == name}
        return Collection(items=items)

    def merge(self, prefix: str, collection) -> None:
        """Given a collection and a prefix, merge its items into the collection."""

        prefix = Key.fromstring(prefix)

        for k, item in collection._items.items():
            self._items[prefix + k] = item


def _prepare_definition(defn, attributes):
    """Given a definiton and a bag of attributes, create an instance of it by passing
    in the applicable attributes into the constructor."""

    names = {a.name for a in attr.fields(defn) if a.metadata[Definition.ATTR_ID]}
    args = {name: attributes[name] for name in names}

    return defn(**args)


@attr.s(auto_attribs=True)
class Definition:
    """A definition is the representation of something that can be mapped onto an
    image."""

    ATTR_ID: ClassVar[str] = "arlunio.attribute"

    def __call__(self, width: int = None, height: int = None, **kwargs):
        args = dict(self.definitions)
        attributes = self.attributes

        try:
            width, height = width
        except TypeError:
            pass

        for name in args:

            if name == "width" and args[name] == inspect.Parameter.empty:
                args[name] = width
                continue

            if name == "height" and args[name] == inspect.Parameter.empty:
                args[name] = height
                continue

            if name in kwargs:
                args[name] = kwargs[name]
                continue

            # Else it must be a definition so let's evaluate it
            defn = _prepare_definition(args[name], attributes)
            args[name] = defn(width, height)

        return self._definition(**args, **self.attribs)

    @property
    def attributes(self):
        """All attributes on this definition"""
        return {
            a.name: getattr(self, a.name)
            for a in attr.fields(self.__class__)
            if Definition.ATTR_ID in a.metadata
        }

    @property
    def attribs(self):
        """All non-inherited attributes on this definition"""
        return {
            a.name: getattr(self, a.name)
            for a in attr.fields(self.__class__)
            if not a.metadata[Definition.ATTR_ID]["inherited"]
        }


def _define_attribute(param: inspect.Parameter) -> attr.Attribute:
    """Given a parameter that represents some definition's attribute, write the
    corresponding attrs instance."""

    args = {"default": param.default, "kw_only": True}
    args["metadata"] = {Definition.ATTR_ID: {"inherited": False}}

    if param.annotation != inspect.Parameter.empty:
        args["type"] = param.annotation
        args["validator"] = [attr.validators.instance_of(param.annotation)]

    return attr.ib(**args)


def _inherit_attributes(defn: Definition, attributes):
    """Given a definition and the attributes for the current definition under
    construction copy over its attributes."""

    # Only look at the fields that represent attributes.
    for attrib in (a for a in attr.fields(defn) if a.metadata[Definition.ATTR_ID]):

        # For now we will skip any attributes that have already been defined.
        # This means that any definitions that share the same attribute name will
        # inherit the same value.
        #
        # This is probably what you would want in most(?) cases, but I'm sure we will
        # need the ability to do something more clever in the future...
        if attrib.name in attributes:
            continue

        metadata = dict(**attrib.metadata)
        metadata[Definition.ATTR_ID] = {"inherited": True}

        # It seems that the best way to copy an attrs attribute from one class to
        # another is to construct a fresh instance, copying over all the fields from
        # the source...
        attributes[attrib.name] = attr.ib(
            converter=attrib.converter,
            default=attrib.default,
            kw_only=attrib.kw_only,
            repr=False,
            metadata=metadata,
            validator=attrib.validator,
        )


def _process_parameters(
    signature: inspect.Signature, attributes: Dict[str, Any]
) -> List[inspect.Parameter]:
    """Ensure that the input parameters for the defintion are well defined.

    First this function will ensure any of the parameters that are not a known "base"
    (e.g. :code:`width` or :code:`height`) are instead a known :code:`Definition`.
    For any definitions that are referenced, their attributes are then exposed to the
    definition under construction.
    """

    # Input parameters are any arguments that are not attributes
    params = [p for p in signature.values() if p.name not in attributes]

    # We will hold a list of definitions we are based on.
    defns = []

    for param in params:

        if param.annotation == inspect.Parameter.empty:
            if param.name in {"width", "height"}:
                continue

            raise TypeError(f"Unknown input '{param.name}'")

        if not issubclass(param.annotation, Definition):
            raise TypeError(
                f"Invalid input '{param.name}', type '{param.annotation.__name__}'"
                " is not a Definition"
            )

        defns.append(param.annotation)

    # Now for each definition, inherit its attributes.
    for defn in defns:
        _inherit_attributes(defn, attributes)

    return params


def definition(f=None):
    """Create a new Definition."""

    def wrapper(defn):

        name = defn.__name__

        KW_ONLY = inspect.Parameter.KEYWORD_ONLY
        signature = inspect.signature(defn).parameters

        attrs = [attr for attr in signature.values() if attr.kind == KW_ONLY]

        attributes = {
            "__doc__": inspect.getdoc(defn),
            "_definition": staticmethod(defn),
        }

        for a in attrs:
            attributes[a.name] = _define_attribute(a)

        params = _process_parameters(signature, attributes)
        attributes["definitions"] = {p.name: p.annotation for p in params}

        return attr.s(type(name, (Definition,), attributes))

    # Allow the decorator to be used with or without a function call
    if f is None:
        return wrapper

    return wrapper(f)



from ._core import Collection, Definition, definition  # noqa: F401
from ._expressions import all, any, clamp, invert, lerp  # noqa: F401
from ._images import Resolutions, colorramp, encode, fill, save  # noqa: F401
from ._version import __version__  # noqa: F401



import json
import os
import re
import typing

from pathlib import Path

import attr
import docutils.nodes as nodes
import docutils.writers as writers

from docutils.io import StringOutput
from sphinx.builders import Builder
from sphinx.util import logging
from sphinx.util.docutils import SphinxTranslator

from .directives import nbtutorial

logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True)
class NotebookCell:
    """Represents a notebook cell."""

    CODE: typing.ClassVar[str] = "code"
    MARKDOWN: typing.ClassVar[str] = "markdown"

    cell_type: str
    execution_count: int = None
    metadata: typing.Dict[str, typing.Any] = attr.Factory(dict)
    outputs: typing.List[typing.Any] = attr.Factory(list)
    source: str = attr.Factory(str)

    @classmethod
    def code(cls):
        """Create a code cell type"""
        return cls(cls.CODE)

    @classmethod
    def markdown(cls):
        """Create a markdown cell"""
        return cls(cls.MARKDOWN)

    @property
    def json(self):
        cell = attr.asdict(self)
        cell["source"] = [line + "\n" for line in cell["source"].split("\n")]

        md_exclude = {"outputs", "execution_count"}

        def exclude(field):
            return self.cell_type == self.MARKDOWN and field in md_exclude

        return {k: v for k, v in cell.items() if not exclude(k)}


@attr.s(auto_attribs=True)
class Notebook:
    """Represents a notebook."""

    nbformat: int = 4
    nbformat_minor: int = 2
    metadata: typing.Dict[str, typing.Any] = attr.Factory(dict)
    cells: typing.List[NotebookCell] = attr.Factory(list)

    @classmethod
    def fromcells(cls, cells: typing.List[NotebookCell]):
        return cls(cells=cells)

    @property
    def json(self):
        notebook = attr.asdict(self)
        cells = [c.json for c in self.cells]

        notebook["cells"] = cells
        return notebook


class NotebookWriter(writers.Writer):
    """A class that walks a doctree converting it into a notebook."""

    def __init__(self, builder):
        super().__init__()
        self.builder = builder

    def translate(self):

        visitor = NotebookTranslator(self.document, self.builder)
        self.document.walkabout(visitor)
        self.output = visitor.astext()


class NotebookTranslator(SphinxTranslator):
    """Converts rst nodes into their output representation."""

    def __init__(self, document, builder):
        super().__init__(document, builder)
        self.cells = []
        self.level = 0

    @property
    def current_cell(self):

        if len(self.cells) == 0:
            return None

        return self.cells[-1]

    def new_cell(self, cell_type: str) -> None:
        current = self.current_cell

        if current is None or current.cell_type != cell_type:
            cell = NotebookCell(cell_type=cell_type)
            self.cells.append(cell)

    def _log_visit(self, node: nodes.Node, surround: str = None) -> None:
        prefix = self.level * " " + "-> "
        name = node.__class__.__name__

        if surround is not None:
            name = surround + name + surround

        logger.debug(prefix + name)

    def _log_departure(self, node: nodes.Node) -> None:
        pass

    def astext(self):
        notebook = Notebook.fromcells(self.cells)
        return json.dumps(notebook.json, indent=2)

    def visit_section(self, node: nodes.section) -> None:
        self.level += 1
        self._log_visit(node)
        self.new_cell(NotebookCell.MARKDOWN)

    def depart_section(self, node: nodes.section) -> None:
        self.level -= 1
        self._log_departure(node)

    def visit_Text(self, node: nodes.Text) -> None:
        self._log_visit(node)

        if self.current_cell.cell_type == NotebookCell.MARKDOWN:
            self.current_cell.source += node.astext()
            return

        source = node.astext()

        if ">>>" not in source:
            self.current_cell.source += source
            return

        pattern = re.compile("^(>>>|\\.\\.\\.) ?")

        def clean_line(line):
            return pattern.sub("", line)

        cleaned_source = "\n".join(
            [clean_line(line) for line in source.split("\n") if pattern.match(line)]
        )

        self.current_cell.source += cleaned_source

    def depart_Text(self, node: nodes.Text) -> None:
        self._log_departure(node)

    def visit_title(self, node: nodes.title) -> None:
        self._log_visit(node)

        title = "#" * self.level
        self.current_cell.source += f"\n{title} "

    def depart_title(self, node: nodes.title) -> None:
        self._log_departure(node)
        self.current_cell.source += "\n"

    def visit_paragraph(self, node: nodes.paragraph) -> None:
        self._log_visit(node)
        self.new_cell(NotebookCell.MARKDOWN)
        self.current_cell.source += "\n"

    def depart_paragraph(self, node: nodes.paragraph) -> None:
        self._log_departure(node)
        self.current_cell.source += "\n"

    def visit_strong(self, node: nodes.strong) -> None:
        self._log_visit(node)
        self.current_cell.source += "**"

    def depart_strong(self, node: nodes.strong) -> None:
        self._log_departure(node)
        self.current_cell.source += "**"

    def visit_literal_block(self, node: nodes.literal_block) -> None:
        self._log_visit(node)
        self.new_cell(NotebookCell.CODE)

    def depart_literal_block(self, node: nodes.literal_block) -> None:
        self._log_departure(node)

    def visit_literal(self, node: nodes.literal) -> None:
        self._log_visit(node)
        self.current_cell.source += "`"

    def depart_literal(self, node: nodes.literal) -> None:
        self.current_cell.source += "`"

    def visit_list_item(self, node: nodes.list_item) -> None:
        self.current_cell.source += "- "

    def visit_target(self, node: nodes.target) -> None:
        pass

    def visit_reference(self, node: nodes.reference) -> None:
        pass

    def depart_reference(self, node: nodes.reference) -> None:
        attrs = node.attributes

        if attrs["internal"]:
            ref = attrs["refuri"].split("#")[0]
            self.current_cell.source += f"({ref})"

    def visit_inline(self, node: nodes.inline) -> None:
        self.current_cell.source += "["

    def depart_inline(self, node: nodes.inline) -> None:
        self.current_cell.source += "]"

    def unknown_visit(self, node: nodes.Node):
        self._log_visit(node, surround="!")

    def unknown_departure(self, node):
        self._log_departure(node)


class NotebookTutorialBuilder(Builder):
    """Builder that can convert static tutorials into an interactive jupyer
    notebook."""

    name = "nbtutorial"
    format = "ipynb"

    def init(self) -> None:
        """Any initialization goes here."""
        logger.debug(f"[nbtutorial]: Outdir is: {self.outdir}")

    def get_outdated_docs(self) -> typing.Union[str, typing.Iterable[str]]:
        """Not too sure what we should do here yet."""

        return ""

    def get_target_uri(self, docname: str, typ: str = None) -> str:
        """Another method to figure out."""

        uri = docname + ".ipynb"

        logger.debug(f"[nbtutorial]: Target URI: {uri}")

        return uri

    def prepare_writing(self, docnames: typing.Set[str]) -> None:
        """A place we can add logic to?"""

        self.docwriter = NotebookWriter(self)

    def write_doc(self, docname: str, doctree: nodes.Node) -> None:
        logger.debug(f"[nbtutorial]: Called on {docname}")

        # Determine if the document represents a tutorial.
        nodes = list(doctree.traverse(condition=nbtutorial))

        if len(nodes) == 0:
            return

        destination = StringOutput(encoding="utf-8")
        self.docwriter.write(doctree, destination)

        base, fname = os.path.split(docname)
        basedir = os.path.join(self.outdir, base)

        if not os.path.exists(basedir):
            os.makedirs(basedir)

        Path(basedir, "__init__.py").touch()
        outfile = os.path.join(basedir, fname + ".ipynb")

        with open(outfile, "w") as f:
            f.write(self.docwriter.output)
